import asyncio
import fcntl
import logging
import os
import re
import socket
import struct
import time
from contextlib import contextmanager

from yeelib.bulbs import Bulb
from yeelib.upnp import SSDPRequest, SimpleServiceDiscoveryProtocol
from .exceptions import YeelightError

__all__ = ('search_bulbs', 'YeelightProtocol')

logger = logging.getLogger('yeelib')

MCAST_PORT = 1982
MCAST_ADDR = SimpleServiceDiscoveryProtocol.MULTICAST_ADDRESS, MCAST_PORT


class MutableBoolean:
    __state = True

    def __bool__(self):
        return self.__state

    def set(self, value):
        self.__state = bool(value)


@asyncio.coroutine
def send_search_broadcast(transport, search_interval=30, _running=True):
    request = SSDPRequest('M-SEARCH', headers=[
        ('HOST', '%s:%s' % MCAST_ADDR),
        ('MAN', '"ssdp:discover"'),
        ('ST', 'wifi_bulb'),
    ])
    while _running:
        request.sendto(transport, MCAST_ADDR)
        yield from asyncio.sleep(search_interval)


class YeelightProtocol(SimpleServiceDiscoveryProtocol):
    excluded_headers = ['DATE', 'EXT', 'SERVER', 'CACHE-CONTROL', 'LOCATION']
    location_patter = r'yeelight://(?P<ip>\d{1,3}(\.\d{1,3}){3}):(?P<port>\d+)'

    def __init__(self, bulbs, bulb_class=Bulb):
        self.bulbs = bulbs
        self.bulb_class = bulb_class

    @classmethod
    def header_to_kwargs(cls, headers):
        headers = dict(headers)
        location = headers.get('Location')
        cache_control = headers.get('Cache-Control', 'max-age=3600')
        headers = {
            k: v for k, v in headers.items()
            if k.upper() not in cls.excluded_headers
        }

        match = re.match(cls.location_patter, location)
        if match is None:
            raise YeelightError('Location does not match: %s' % location)
        ip = match.groupdict()['ip']
        port = match.groupdict()['port']

        max_age = int(cache_control.rsplit('=', 1)[-1])

        kwargs = dict(ip=ip, port=port, status_refresh_interv=max_age)
        kwargs.update(headers)
        return kwargs

    def request_received(self, request):
        if request.method == 'M-SEARCH':
            return
        self.register_bulb(**self.header_to_kwargs(request.headers))

    def response_received(self, response):
        self.register_bulb(**self.header_to_kwargs(response.headers))

    def register_bulb(self, **kwargs):
        idx = kwargs['id']
        if idx not in self.bulbs:
            self.bulbs[idx] = self.bulb_class(**kwargs)
        else:
            self.bulbs[idx].last_seen = time.time()


@contextmanager
def _unicast_socket():
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as ucast_socket:
        ucast_socket.bind(('', MCAST_PORT))
        fcntl.fcntl(ucast_socket, fcntl.F_SETFL, os.O_NONBLOCK)
        group = socket.inet_aton(
            SimpleServiceDiscoveryProtocol.MULTICAST_ADDRESS)
        mreq = struct.pack("4sl", group, socket.INADDR_ANY)
        ucast_socket.setsockopt(
            socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        yield ucast_socket


@contextmanager
def search_bulbs(bulbs=None, bulb_class=Bulb, loop=None):
    if bulbs is None:
        bulbs = {}
    if loop is None:
        loop = asyncio.get_event_loop()
    multicast_connection = loop.create_datagram_endpoint(
        lambda: YeelightProtocol(bulbs, bulb_class),
        family=socket.AF_INET)
    mcast_transport, _ = loop.run_until_complete(multicast_connection)
    running = MutableBoolean()
    loop.create_task(send_search_broadcast(mcast_transport, _running=running))

    with _unicast_socket() as ucast_socket:
        unicast_connection = loop.create_datagram_endpoint(
            lambda: YeelightProtocol(bulbs, bulb_class),
            sock=ucast_socket)
        ucast_transport, _ = loop.run_until_complete(unicast_connection)
        yield bulbs
    running.set(False)
    ucast_transport.close()
    mcast_transport.close()



import asyncio
import email.parser
import errno
import logging

__all__ = ('SSDPRequest', 'SSDPResponse', 'SimpleServiceDiscoveryProtocol')

logger = logging.getLogger('yeelib')


class SSDPMessage:
    """Simplified HTTP message to serve as a SSDP message."""

    def __init__(self, version='HTTP/1.1', headers=None):
        if headers is None:
            headers = []
        elif isinstance(headers, dict):
            headers = headers.items()

        self.version = version
        self.headers = list(headers)

    @classmethod
    def parse(cls, msg):
        """
        Parse message from string.

        Args:
            msg (str): Message string.

        Returns:
            SSDPMessage: Message parsed from string.

        """
        raise NotImplementedError()

    @classmethod
    def parse_headers(cls, msg):
        """
        Parse HTTP headers.

        Args:
            msg (str): HTTP message.

        Returns:
            (List[Tuple[str, str]): List of header tuples.

        """
        return list(email.parser.Parser().parsestr(msg).items())


class SSDPResponse(SSDPMessage):
    """Simple Service Discovery Protocol (SSDP) response."""

    def __init__(self, status_code, reason, **kwargs):
        self.status_code = int(status_code)
        self.reason = reason
        super().__init__(**kwargs)

    @classmethod
    def parse(cls, msg):
        lines = msg.splitlines()
        version, status_code, reason = lines[0].split()
        headers = cls.parse_headers('\r\n'.join(lines[1:]))
        return cls(version=version, status_code=status_code,
                   reason=reason, headers=headers)


class SSDPRequest(SSDPMessage):
    """Simple Service Discovery Protocol (SSDP) request."""

    def __init__(self, method, uri='*', version='HTTP/1.1', headers=None):
        self.method = method
        self.uri = uri
        super().__init__(version=version, headers=headers)

    @classmethod
    def parse(cls, msg):
        lines = msg.splitlines()
        method, uri, version = lines[0].split()
        headers = cls.parse_headers('\r\n'.join(lines[1:]))
        return cls(version=version, uri=uri, method=method, headers=headers)

    def sendto(self, transport, addr):
        """
        Send request to a given address via given transport.

        Args:
            transport (asyncio.DatagramTransport):
                Write transport to send the message on.
            addr (Tuple[str, int]):
                IP address and port pair to send the message to.

        """
        msg = bytes(self) + b'\r\n'
        logger.debug("%s:%s < %s", *(addr + (self,)))
        transport.sendto(msg, addr)

    def __str__(self):
        lines = list()
        lines.append(' '.join(
            [self.method, self.uri, self.version]
        ))
        for header in self.headers:
            lines.append('%s: %s' % header)
        return '\n'.join(lines)

    def __bytes__(self):
        return self.__str__().encode().replace(b'\n', b'\r\n')


class SimpleServiceDiscoveryProtocol(asyncio.DatagramProtocol):
    """
    Simple Service Discovery Protocol (SSDP).

    SSDP is part of UPnP protocol stack. For more information see:
    https://en.wikipedia.org/wiki/Simple_Service_Discovery_Protocol
    """

    MULTICAST_ADDRESS = '239.255.255.250'

    def datagram_received(self, data, addr):
        data = data.decode()
        logger.debug("%s:%s > %s", *(addr + (data,)))

        if data.startswith('HTTP/'):
            self.response_received(SSDPResponse.parse(data))
        else:
            self.request_received(SSDPRequest.parse(data))

    def response_received(self, response):
        """
        Called when some response is received.

        Args:
            response (SSDPResponse): Received response.

        """
        raise NotImplementedError()

    def request_received(self, request):
        """
        Called when some request is received.

        Args:
            request (SSDPRequest): Received request.

        """
        raise NotImplementedError()

    def error_received(self, exc):
        if exc == errno.EAGAIN or exc == errno.EWOULDBLOCK:
            logger.error('Error received: %s', exc)
        else:
            raise IOError("Unexpected connection error") from exc

    def connection_lost(self, exc):
        logger.exception("Socket closed, stop the event loop", exc_info=exc)
        loop = asyncio.get_event_loop()
        loop.stop()



import pytest

from yeelib.upnp import SSDPMessage, SSDPResponse, SSDPRequest
from . import fixtures


class TestSSDPMessage:
    def test_headers_copy(self):
        headers = [('Cache-Control', 'max-age=3600')]
        msg = SSDPMessage(headers=headers)
        assert msg.headers == headers
        assert msg.headers is not headers

    def test_headers_dict(self):
        headers = {'Cache-Control': 'max-age=3600'}
        msg = SSDPMessage(headers=headers)
        assert msg.headers == [('Cache-Control', 'max-age=3600')]

    def test_headers_none(self):
        msg = SSDPMessage(headers=None)
        assert msg.headers == []

    def test_parse(self):
        with pytest.raises(NotImplementedError):
            SSDPMessage.parse('')

    def test_parse_headers(self):
        headers = SSDPMessage.parse_headers('Cache-Control: max-age=3600')
        assert headers == [('Cache-Control', 'max-age=3600')]


class TestSSDPResponse:
    def test_parse(self):
        response = SSDPResponse.parse(fixtures.response.decode())
        assert response.status_code == 200
        assert response.reason == 'OK'


class TestSSDPRequest:
    def test_parse(self):
        request = SSDPRequest.parse(fixtures.request.decode())
        assert request.method == 'NOTIFY'
        assert request.uri == '*'

    def test_str(self):
        request = SSDPRequest('NOTIFY', '*', headers=[('Cache-Control', 'max-age=3600')])
        assert str(request) == (
            'NOTIFY * HTTP/1.1\n'
            'Cache-Control: max-age=3600'
        )

    def test_bytes(self):
        request = SSDPRequest.parse(fixtures.request.decode())
        assert bytes(request) == fixtures.request
